import datetime
import random
import string
import codecs
import subprocess
import tempfile

from metaindex.ocr import LANGUAGES


DATE_FORMAT = '%Y-%m-%d'
DATE_TIME_SEPARATOR = 'T'
DT_FORMAT = '{}{}%H:%M:%S'.format(DATE_FORMAT, DATE_TIME_SEPARATOR)


def random_string(length=32, characters=None):
    if characters is None:
        characters = string.ascii_letters + string.digits + '.-_+/='
    return ''.join(random.choices(characters, k=length))


def random_id():
    return random_string(24, string.ascii_lowercase + string.digits)


def human_readable_size(size):
    rem = size
    units = ['', 'K', 'M', 'G', 'T']
    unit = 0

    while rem >= 1024:
        rem /= 1024.
        unit += 1

    return str(round(rem, 2)) + ' ' + units[unit] + 'b'


def date_from_str(text):
    return datetime.datetime.strptime(text, DATE_FORMAT).date()


def readable_code(code, sep=' '):
    groupsize = 2
    for gs in [4, 3, 5]:
        if len(code) % gs == 0:
            groupsize = gs
            break

    result = ''
    for idx, c in enumerate(code):
        if idx > 0 and idx % groupsize == 0:
            result += sep
        result += c
    return result


def parse_duration(text):
    if not text.startswith('P'):
        raise ValueError('Not a ISO 8601 duration')

    pos = 1
    value = ''
    days = 0
    seconds = 0

    while pos < len(text):
        if text[pos] == 'T':
            pass
        elif text[pos] in 'YDWHMS':
            if text[pos] == 'Y':
                days += int(value)*365  # this is weird
            elif text[pos] == 'W':
                days += int(value)*7
            elif text[pos] == 'D':
                days += int(value)
            elif text[pos] == 'H':
                seconds += int(value)*3600
            elif text[pos] == 'M':
                seconds += int(value)*60
            elif text[pos] == 'S':
                seconds += int(value)
            value = ''
        elif text[pos] in string.digits:
            value += text[pos]
        else:
            raise ValueError(f"{text} is not a ISO 8601 duration. First problem at '{text[pos]}'.")
        pos += 1

    return datetime.timedelta(days=days, seconds=seconds)


def parse_size(text, fallback=0):
    text = text.lower().strip()
    factor = 1
    try:
        for postfix, f_ in [('g', 1024*1024*1024), ('m', 1024*1024), ('k', 1024), ('b', 1)]:
            if text.endswith(postfix):
                text = text[:-1]
                factor = f_
                break
        return int(text)*factor
    except ValueError:
        return fallback


def to_utf8(raw):
    if isinstance(raw, str):
        return raw
    encoding = None
    strip_from = 1

    if raw.startswith(codecs.BOM_UTF8):
        encoding = 'utf-8'
    elif raw.startswith(codecs.BOM_UTF16_BE):
        encoding = 'utf-16-be'
    elif raw.startswith(codecs.BOM_UTF16_LE):
        encoding = 'utf-16-le'
    elif raw.startswith(codecs.BOM_UTF32_BE):
        encoding = 'utf-32-be'
    elif raw.startswith(codecs.BOM_UTF32_LE):
        encoding = 'utf-32-le'
    else:
        # just guessing
        encoding = 'utf-8'
        strip_from = 0

    try:
        text = str(raw, encoding=encoding).strip()
        return text[strip_from:]  # drop the BOM
    except UnicodeError:
        pass
    return None


def get_language_from_filename(fn):
    parts = fn.stem.rsplit('-', 1)
    if len(parts) != 2:
        return None
    lang = LANGUAGES.get(parts[1], parts[1])
    if lang not in LANGUAGES.values():
        return None
    return lang


def edit_fulltext_externally(version, program):
    to_edit = "\n".join([part.fulltext for part in version.parts])
    edited = to_edit

    with tempfile.NamedTemporaryFile('w+t', encoding='utf-8', newline="\n") as tf:
        tf.write(to_edit)
        tf.flush()
        subprocess.run(program + [tf.name])
        tf.seek(0)
        edited = tf.read()

    if edited != to_edit:
        for part in version.parts:
            part.technical.fulltext = ''
        version.parts[0].technical.fulltext = edited



DUBLINCORE = {'contributor', 'coverage', 'creator',
              'date', 'description', 'format', 'identifier',
              'language', 'publisher', 'relation', 'rights',
              'source', 'subject', 'title', 'type'}
TAGGING = {'keywords', 'tags', 'categories'}
ALL_METADATA = DUBLINCORE.union(TAGGING)




"""Package for Cabinet and related classes"""
import logging
import shutil
import subprocess
import mimetypes
import tempfile
from pathlib import Path

from pypdf import PdfWriter

import metaindex.shared
from metaindex import yaml
from metaindex import indexers
from metaindex import indexer
from metaindex import CacheEntry

from filecabinet import utils


MIMETYPE_PDF = 'application/pdf'


class LanguageFromFilenameIndexer(metaindex.indexer.IndexerBase):
    """An indexer that extracts the language of the file from its name"""
    NAME = 'filecabinet_filenamelanguage'
    ACCEPTS = '*'
    PREFIX = 'extra.'
    ORDER = metaindex.indexer.Order.EARLY

    def run(self, path, metadata, _):
        language = utils.get_language_from_filename(path)
        if language is not None:
            metadata.add(self.PREFIX + 'language', language)
            other = utils.LANGUAGES.get(language, language)
            if other != language:
                metadata.add(self.PREFIX + 'language', other)


class Document(CacheEntry):
    """Representing a document in a filecabinet"""
    def __init__(self, *args, cabinet=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.cabinet = cabinet

    @classmethod
    def wrap(cls, cabinet, entry):
        """Create a new Document based on ``entry``
        """
        if isinstance(entry, Document):
            entry.cabinet = cabinet
            return entry

        that = cls(entry.path, cabinet=cabinet)
        that.metadata = entry.metadata
        that.last_modified = entry.last_modified
        that.mimetype = entry.mimetype
        that.rel_path = entry.rel_path
        that.storage_label = entry.storage_label
        return that

    def copy(self):
        that = super().copy()
        that.cabinet = self.cabinet
        return that

    def is_new(self):
        return any(v == 'new' for v in self.get('extra.tag'))


class Cabinet:
    """Represents a filecabinet as a container of documents"""
    def __init__(self, name, manager):
        self.manager = manager
        self.config = manager.config
        self.name = name
        self.description = self.config[name].get('description', name)
        self.basedir = self.config.path(name, 'path')
        self.inbox = self.basedir / self.config[name].get('inbox', 'inbox')
        self.documents = self.basedir / self.config[name].get('documents', 'documents')

        self.documents.mkdir(exist_ok=True, parents=True)
        self.inbox.mkdir(exist_ok=True, parents=True)

    def next_doc_id(self):
        while True:
            fullid = utils.random_id()
            prefix = fullid[:4]

            if (self.documents / prefix / fullid).is_dir():
                continue

            return (prefix, fullid)

    def pickup(self, session):
        files = list(self.inbox.iterdir())

        for fn in files[:]:
            if not fn.is_dir():
                continue
            files.remove(fn)
            merged = self.merge_files(fn)
            if merged is not None:
                files.append(merged)

        if self.manager.ocr is not None:
            # convert picture files to OCR'd PDFs
            ## find picture files first
            images = []
            for fn in files[:]:
                mimetype, _ = mimetypes.guess_type(fn, strict=False)
                if mimetype is None:
                    continue
                if not mimetype.startswith('image/'):
                    continue
                images.append(fn)
                files.remove(fn)

            ## process each image file
            for fn in images:
                newfn = self.manager.process_image(fn)
                if newfn is not None:
                    files.append(newfn)
                    try:
                        fn.unlink()
                    except OSError as exc:
                        logging.error(f"Could not clean up '{fn}': {exc}")

        self.index_new_files(session, files)

    def index_new_files(self, session, files):
        """Index the given files and move them to this cabinet's 'document' folder"""
        generator = indexer.index_files(files, self.manager.client_config, 1, None, True, {}, {})

        for result in generator:
            if not result.success:
                logging.warning(f"Unknown file: {result.filename}")
                continue

            result.info.last_modified = metaindex.shared.get_last_modified(result.filename)
            info = result.info

            # new document ID
            prefix, docid = self.next_doc_id()
            docdir = self.documents / prefix / docid
            docdir.mkdir(parents=True)
            docpath = docdir / (docid + info.path.suffix)

            # move to new location
            shutil.move(result.filename, docpath)
            info.path = docpath

            # mark as 'new'
            info.add('extra.tag', 'new')

            # index metadata in cache
            session.cache.reduce(info)
            session.cache.insert(info)

            # split off fulltext
            fulltext = ("\n".join([str(v) for v in info.get('extra.fulltext')])).strip()

            # create the sidecar file
            self.save_metadata(info)

            # write out full text, if there was any
            if len(fulltext) > 0:
                text = docdir / (docid + ".txt")
                text.write_text(fulltext, encoding='utf-8')

    def save_metadata(self, document):
        """Save the metadata of 'document' to its respective sidecar file"""
        docdir = document.path.parent
        docid = document.path.stem
        assert docdir.stem == docid

        to_save = document.copy()
        to_save.delete('extra.fulltext')

        # create sidecar file
        sidecar = docdir / (docid + ".yaml")
        yaml.store(to_save, sidecar)

    def merge_files(self, dirpath):
        """Try and merge all files in ``dirpath`` into one PDF at ``dirpath/../dirpath.pdf``

        Returns the full path to the merged file on success, otherwise ``None``
        """
        assert dirpath.is_relative_to(self.inbox)
        lang = utils.get_language_from_filename(dirpath)

        parts = list(dirpath.iterdir())
        parts.sort()
        mimes = {}
        # if there's anything that's neither image nor PDF in these files, we're out
        for part in parts:
            mimetype, _ = mimetypes.guess_type(part, strict=False)
            mimes[part] = mimetype
            if mimetype is None:
                return None
            if mimetype == MIMETYPE_PDF:
                # PDFs can be merged
                continue
            if not mimetype.startswith('image/'):
                # some file, not an image, not a PDF
                logging.error(f"Can only merge images and PDFs ({dirpath})")
                return None

        # convert all images to PDF and OCR them
        converted = []
        for part in parts:
            mimetype = mimes[part]
            if mimetype == MIMETYPE_PDF:
                converted.append(part)
                continue

            if self.manager.ocr is None:
                logging.error("Tesseract OCR is disabled or not available. "
                              f"Can not merge files in '{dirpath}' meaningfully.")
                return None

            pdffile = self.manager.process_image(part, lang)
            if pdffile is None:
                logging.error(f"Could not convert {part} to PDF")
                return None
            try:
                part.unlink()
            except OSError as exc:
                logging.error(f"Could not clean up converted file '{part}': {exc}")
            converted.append(pdffile)

        # time to merge all pages in `converted`
        with tempfile.NamedTemporaryFile(suffix='.pdf',
                                         dir=dirpath.parent,
                                         delete=False) as tmpfd:
            targetfile = Path(tmpfd.name)

        writer = PdfWriter()
        for part in converted:
            writer.append(part)
        writer.write(targetfile)
        writer.close()

        try:
            shutil.rmtree(dirpath)
        except OSError as exc:
            logging.error(f"Could not delete {dirpath}: {exc}")
        return targetfile

    def files(self):
        """Iterate over all document files in this cabinet"""
        paths = sum([list(d.iterdir())
                     for d in self.documents.iterdir()], start=[])

        for path in paths:
            candidates = [fn for fn in path.iterdir() if fn.suffix != '.yaml']
            if len(candidates) == 0:
                logging.info(f"Abandoned folder {path}")
                continue
            if len(candidates) == 1:
                yield candidates[0]
                continue
            candidates = [c for c in candidates if c.suffix != '.txt']
            assert len(candidates) > 0
            yield candidates[0]

    def reindex(self, session, files=None):
        """Re-index the given files

        If no files were selected, re-index the entire cabinet
        """
        if files is None:
            files = list(self.files())

        if not isinstance(files, (list, tuple, set)):
            files = [files]

        files = [fn for fn in files
                 if fn.is_relative_to(self.documents)]

        for filepath in files:
            sidecar = filepath.parent / (filepath.stem + '.yaml')
            fulltext = filepath.parent / (filepath.stem + '.txt')

            entry = yaml.get(sidecar)
            entry.path = filepath

            if fulltext.is_file():
                entry.add('extra.fulltext', fulltext.read_text(encoding='utf-8'))

            session.cache.insert(entry)

    def update(self, session, documents):
        """Update the metadata information of these documents"""
        for document in documents:
            if not document.path.is_relative_to(self.documents):
                continue
            self.save_metadata(document)
            session.cache.insert(document)



import argparse
import logging
import sys
import re
import mimetypes
from pathlib import Path

from metaindex import indexers
from metaindex import indexer

from filecabinet.configuration import get_configuration, CONFIGFILE
from filecabinet.manager import Manager
from filecabinet.shell import Shell


CABINET_ID_RE = re.compile(r'^[\w][^\t\n\r\f\v]*$')


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config',
                        default=None,
                        type=str,
                        help=f'The config file you wish to use. Defaults to {CONFIGFILE}')
    parser.add_argument('-l', '--log-level',
                        choices=['debug', 'info', 'warning', 'error', 'fatal'],
                        default=None,
                        help='The loglevel.')
    
    subparsers = parser.add_subparsers(dest='command')

    serviceparser = subparsers.add_parser('service', help='Maintenance tasks')
    serviceparser.add_argument('action',
                               choices=['stop', 'compact', 'start'],
                               help="What task to perform. Options are: 'stop' "
                                    "to stop the backend server, 'start' to "
                                    "start the backend server, 'compact' to "
                                    "compact the backend database.")

    subparsers.add_parser('shell', help='Start the filecabinet shell')

    listparser = subparsers.add_parser('list', help='List various things')
    listparser.add_argument('what',
                            choices=['cabinets', 'documents', 'new'],
                            help="What to list. Options are: %(choices)s.")

    findparser = subparsers.add_parser('find', help="Find documents")
    findparser.add_argument('query',
                            nargs='*',
                            type=str,
                            help="The search query to run")

    rulesparser = subparsers.add_parser('test-rules', help="Test your rules on a file")
    rulesparser.add_argument('-t', '--text',
                             action="store_true",
                             default=False,
                             help="Print the full text that was found.")
    rulesparser.add_argument('file',
                             type=str,
                             help="The file to run the rules on")

    initparser = subparsers.add_parser('init', help="Initialize a new file cabinet")
    initparser.add_argument('-d', '--description',
                            default='',
                            type=str,
                            help="Optional single line description.")
    initparser.add_argument('-n', '--name',
                            default='cabinet',
                            type=str,
                            help="Unique name for this file cabinet. Defaults to %(default)s.")
    initparser.add_argument('path',
                            type=str,
                            help="Path where the documents should be stored.")

    subparsers.add_parser('pickup', help='Pick up files from incoming')

    reindexparser = subparsers.add_parser('reindex', help='Rebuild index')
    reindexparser.add_argument('-c', '--cabinet',
                               default=None,
                               type=str,
                               help="The cabinet to reindex. If there is only "
                                    "one cabinet, this can be left empty.")

    args = parser.parse_args(sys.argv[1:])
    if args.command not in ['service', 'pickup', 'list', 'find', 'test-rules', 'reindex', 'init', 'shell']:
        parser.print_help()

    return args


def error(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def run():
    logging.basicConfig(level=logging.WARNING)
    args = parse_args()

    conf = get_configuration(args.config)

    level = args.log_level if args.log_level is not None else conf['General']['loglevel'].upper()
    logging.getLogger().setLevel(level.upper())

    manager = Manager(conf)

    if args.command == 'init':
        sys.exit(create_new_cabinet(manager, args))

    elif args.command == 'test-rules':
        sys.exit(test_rules(manager, args))

    elif len(manager.cabinets) == 0:
        logging.warning(f"There are no cabinets configured.")
        sys.exit(-1)

    manager.launch_server()

    if args.command == 'service':
        session = manager.new_session()

        if args.action == 'stop':
            session.cache.shutdown()

        if args.action == 'compact':
            session.cache.compact()

        if args.action == 'start':
            pass  # already done

    elif args.command == 'pickup':
        session = manager.new_session()
        session.pickup()

    elif args.command == 'find':
        query = " ".join(args.query)
        session = manager.new_session()
        for result in session.find(query):
            print(result)

    elif args.command == 'reindex':
        session = manager.new_session()
        session.reindex()

    elif args.command == 'list':
        if args.what == 'cabinets':
            for cabinet in manager.cabinets:
                print(f"{cabinet.name}\t{cabinet.basedir}")

        if args.what == 'new':
            session = manager.new_session()
            for doc in session.find('tag:new'):
                print(doc.path)

        if args.what == 'documents':
            raise NotImplementedError("Not gonna do that just now")

    elif args.command == 'shell':
        Shell(manager.new_session()).run()

    sys.exit(0)


def create_new_cabinet(manager, args):
    # 1. check if the path to the config file exists
    confpath = manager.config.filepath
    if confpath is None:
        confpath = CONFIGFILE
    try:
        confpath.parent.mkdir(parents=True, exist_ok=True)
    except OSError as exc:
        error(f"Could not create path to configuration file '{confpath.parent}': {exc}")
        return -2

    # 2. check that the path to the new cabinet does not exist
    cabinetpath = Path(args.path).expanduser().absolute()
    if cabinetpath.exists() and any(cabinetpath.iterdir()):
        error(f"Path for new file cabinet '{cabinetpath}' exists and is not empty!")
        cabinet = manager.path_cabinet.get(cabinetpath)
        if cabinet is not None:
            error(f"Cabinet {cabinet.name} is defined at that location.")
        return -3

    # 3. touch config file (to test we can write)
    try:
        confpath.touch()
    except OSError as exc:
        error(f"Cannot write to '{confpath}': {exc}")
        return -4

    # 4. determine a good cabinet id
    cabinetcounter = ''
    cabinetbasename = args.name
    if not CABINET_ID_RE.match(cabinetbasename):
        error(f"'{args.name}' is not a valid name for a file cabinet. "
              "Please do not use new-line or control characters.")
        return -6

    while True:
        cabinetid = f'{cabinetbasename}{cabinetcounter}'
        if cabinetid in manager.config:
            if cabinetcounter == '':
                cabinetcounter = 1
            else:
                cabinetcounter += 1
            continue
        break

    # 5. create the cabinet path
    inboxname = 'inbox'  # TODO allow configuration
    documentsname = 'documents'  # TODO allow configuration
    try:
        cabinetpath.mkdir(parents=True, exist_ok=True)
        (cabinetpath / inboxname).mkdir()
        (cabinetpath / documentsname).mkdir()
    except OSError as exc:
        error(f"Could not create paths for new file cabinet: {exc}")
        return -5


    # 6. add the cabinet to the config
    name = ''
    if len(args.name.strip()) > 0:
        name = f'name = {args.name.strip()}'
    text = ''
    if confpath.exists():
        text = confpath.read_text(encoding='utf-8')
        text += "\n\n"
    text += f"""[{cabinetid}]
path = {cabinetpath}
inbox = {inboxname}
documents = {documentsname}
{name}
"""
    confpath.write_text(text, encoding='utf-8')

    return 0


def test_rules(manager, args):
    # 1. Check that rule file are defined and exist
    if 'Rules' not in manager.config or len(manager.config['Rules']) == 0:
        error("No rule files defined in your configuration file {manage.config.filepath}.")
        return -1

    indexer.logger.setup(logging.WARNING)

    missing_rule_files = [manager.config.path('Rules', name)
                          for name in manager.config['Rules']
                          if not manager.config.path('Rules', name).is_file()]
    if len(missing_rule_files) > 0:
        logging.warning("These rule files are defined but don't actually exist: %s" ,
                        ", ".join(str(p) for p in missing_rule_files))
    if len(missing_rule_files) == len(manager.config['Rules']):
        error("No rule files exist on disk.")
        return -2

    # 2. extract fulltext from file, OCR if required
    sourcefile = Path(args.file).expanduser()
    to_index = [sourcefile]
    delete_me = None
    mime, _ = mimetypes.guess_type(sourcefile, strict=False)
    if mime is not None and mime.startswith('image/'):
        delete_me = manager.process_image(sourcefile)
        to_index = [delete_me]

    results = list(indexer.index_files(to_index, manager.client_config, 1, None, True, {}, {}))
    for result in results:
        if args.text:
            fulltext = "\n".join(value.raw_value
                                 for tag, value in result.info
                                 if tag.endswith('.fulltext'))
            print(fulltext)
        for tag, value in result.info:
            if not tag.startswith('rules.'):
                continue
            print(f"{tag}: {value}")

    if delete_me is not None:
        try:
            delete_me.unlink()
        except OSError as exc:
            error(f"Could not delete temporary file '{delete_me}': {exc}")

    return 0



import setuptools

from filecabinet import version


with open('README.md', encoding='utf-8') as fd:
    long_description = fd.read()

with open('LICENSE', encoding='utf-8') as fd:
    licensetext = fd.read()


setuptools.setup(name='filecabinet',
      version=version.VERSION,
      description="A local, offline document archive",
      long_description=long_description,
      long_description_content_type='text/markdown',
      url="https://vonshednob.cc/filecabinet",
      author="R",
      author_email="dev+filecabinet-this-is-spam@vonshednob.cc",
      license=licensetext,
      entry_points={'console_scripts': ['filecabinet=filecabinet.main:run']},
      packages=['filecabinet',
                ],
      package_data={},
      data_files=[],
      install_requires=['pyyaml', 'pypdf', 'Pillow', 'metaindex >= 2.2.2'],
      python_requires='>=3.7',
      classifiers=['Development Status :: 3 - Alpha',
                   'License :: OSI Approved :: BSD License',
                   'Programming Language :: Python :: 3',
                   ])
